/*------------------------------------------------------------------------------
 Copyright:      Radig Ulrich, (Ralf Figge)  mailto: mail@ulrichradig.de
 Author:         Radig Ulrich, reworked by Wolfgang Scherr
 Remarks:        Rework: massiv berarbeitet, jetzt viel flexibler
 known Problems: none
 Version:        12.02.2009
 Description:    Programm zur Ansteuerung eines Standard LCD
				 (HD44870),(SED1278F) und kompatible
------------------------------------------------------------------------------*/

#include "config.h"

#if USE_LCD

	#include <avr/io.h>
	#include <stdlib.h>
	#include <util/delay.h>
	
#if !USE_LCD_STREAM
	#include <avr/pgmspace.h>
	#include <stdarg.h>
	#include <ctype.h>
	#include <string.h>
#else
	#include <stdio.h>
	#if LCD_SER
		#warning "Serial LCD IF: LCD streaming does not scroll!"
	#endif
#endif

	#include "lcd.h"

#ifdef COL20
const char LINE_ADDRESS[] ={0x80,0xC0,0x94,0xD4}; //1x20, 2x20, 4x20
#else
const char LINE_ADDRESS[] ={0x80,0xC0,0x90,0xD0}; //1x16, 2x16, 4x16
#endif

#if LCD_SER
	// Zustand merken fr maskiertes Schreiben.
	unsigned char lcd_shiftreg;
#endif

//------------------------------------------------------------------------------
void lcd_init (void)
{
	//To avoid glitches on E, reset it first before switching to output.
	//(just to be sure when coming from an unexpected RESET)
	LCD_EL;
	DDR_LCD_E |= (1<<LCD_E);
	
	#if LCD_PAR
		//Set Port Direction Register to Output for LCD Databus and Control
		DDR_LCD_DATA |= LCD_DATA_MASK;
		DDR_LCD_RW |= (1<<LCD_RW);
		DDR_LCD_RS |= (1<<LCD_RS);
		#if USE_LCD_BACKLIGHT
			DDR_LCD_BL |= (1<<LCD_BL);
		#endif
	#elif LCD_SER
		DDR_LCD_DATA |= (1<<LCD_DATA);
		DDR_LCD_CLOCK |= (1<<LCD_CLOCK);
		// Zustand initialisieren
		lcd_shiftreg = 0;
		shift_data(lcd_shiftreg);
	#else
		#error "Keine LCD Konfiguration gefunden, breche ab!"
	#endif

	//Wait a short Time after Power on	
	_delay_ms(1);
		
	for (unsigned char a=0;a<3;a++)
		{
		lcd_write (0x22,0);	//Init in 4 Bit Mode
		lcd_write (0x80,0);	//Set DD-Ram Adresse = 0
		}
	#ifdef ONE_LINES
		lcd_write (0x20,0);	//1 Zeilen
	#else
		lcd_write (0x28,0);	//mehrere Zeilen
	#endif
	
	lcd_write (0x0C,0);	//Display On
}

//------------------------------------------------------------------------------
void lcd_write (char data, char datareg)
{
	#if LCD_PAR	&& LCD_PAR_READ
		unsigned int timeout = 0;
		//Wait until LCD is ready
		while ((lcd_read(0)&(1<<BUSYBIT)) > 0 && timeout++ < 500) {};
	#else
		_delay_us(10);
	#endif
	
	//Soll ins Steuer oder Datenregister geschrieben werden?
	if (datareg)
		LCD_RSH; //RS = 1 Dataregister		
	else
		LCD_RSL; //RS = 0 Steuerregister

	//Schreibsignal setzen
	LCD_RWL;	//Zum Schreiben RW-Pin = Low
		
	//Schreiben der 1. 4Bit an das LCD Display
	LCD_DATA_WRITEH(data);	//High-Nibble
	LCD_EH;
	LCD_EL;

	//Schreiben der 2. 4Bit an das LCD Display	
	LCD_DATA_WRITEL(data);	//Low-Nibble
	LCD_EH;
	LCD_EL;

#if (!LCD_PAR_READ) || LCD_SER
	// in case of a command, always give the LCD time in case we
	// don't read the status!
	if (!datareg)
		_delay_ms(10);	// Commands take longer
	else
		_delay_us(300); // Data write should be faster
#endif
}

#if LCD_PAR	&& LCD_PAR_READ
	char lcd_read (char datareg)
	{
		char Data;

		//Set Port Direction Register to Input for LCD Databus
		DDR_LCD_DATA &= ~LCD_DATA_MASK;

		//Soll vom Steuer oder Datenregister gelesen werden?
		if (datareg)
			LCD_RSH; //RS = 1 Dataregister		
		else
			LCD_RSL; //RS = 0 Steuerregister

		//Lesesignal setzen
		LCD_RWH;	//Zum Lesen RW-Pin = High

		LCD_EH; //Daten stehen an wenn Enable = High
		Data = LCD_DATA_READH; //High-Nibble
		LCD_EL;

		LCD_EH; //Daten stehen an wenn Enable = High
		Data |= LCD_DATA_READL; //Low-Nibble
		LCD_EL;	

		_delay_us(1); //Wir bleiben auf der sicheren Seite :-)

		//Set Port Direction Register to Output for LCD Databus
		DDR_LCD_DATA |= LCD_DATA_MASK;

		return(Data);
	}
#endif

#if LCD_SER
	void shift_data(char val) {
		unsigned char i;
		
		// Wir schieben mit dem MSB zuerst
        for (i=0; i<8; i++)
        {
			if (val&0x80)
				PORT_LCD_DATA |= (1<<LCD_DATA); //Data PIN (High)
			else
				PORT_LCD_DATA &= ~(1<<LCD_DATA); //Data PIN (LOW)
			val <<= 1;
			PORT_LCD_CLOCK |= (1<<LCD_CLOCK); //Clock PIN (High)
			asm("nop");
			asm("nop");
			PORT_LCD_CLOCK &= ~(1<<LCD_CLOCK); //Clock PIN (Low)
        }
	}
#endif

//------------------------------------------------------------------------------
unsigned char lcd_cx;
unsigned char lcd_cy;

void lcd_clear (void)
{
	lcd_cx=0;
	lcd_cy=0;
	lcd_write (1,0); //Clear Display
	lcd_write (0x80,0);	//Set DD-Ram Adresse = 0
}
//------------------------------------------------------------------------------
#if !LCD_SER
void lcd_scroll(void)
{   unsigned int timeout = 0;

	unsigned char zeile,i,j=0;
	#ifdef COL20
		unsigned char buf[20];
	#else
		unsigned char buf[16];
	#endif
	
	#ifdef TWO_LINES
	    for(j=0;j<1;j++) {
	#endif
	#ifdef FOUR_LINES
	    for(j=0;j<3;j++) {
	#endif
			zeile = LINE_ADDRESS[j+1];
			lcd_write(zeile,0);
			for(i=0;i<sizeof(buf);i++) {
				//Wait until LCD is ready
				while ((lcd_read(0)&(1<<BUSYBIT)) > 0 && timeout++ < 500) {};
				buf[i]=lcd_read(1);
			}
			zeile = LINE_ADDRESS[j];
			lcd_write(zeile,0);
			for(i=0;i<sizeof(buf);i++) {
				//Wait until LCD is ready
				while ((lcd_read(0)&(1<<BUSYBIT)) > 0 && timeout++ < 500) {};
				lcd_write(buf[i],1);
			}
		}
		zeile = LINE_ADDRESS[j];
		lcd_write(zeile,0);
		for(i=0;i<sizeof(buf);i++) {
			//Wait until LCD is ready
			while ((lcd_read(0)&(1<<BUSYBIT)) > 0 && timeout++ < 500) {};
			lcd_write(' ',1);
		}
}
#endif
//------------------------------------------------------------------------------
void lcd_pos (unsigned char zeile,unsigned char spalte)
{
	#ifdef COL20
		while (spalte>=20) { spalte-=20; zeile++; }
	#else
		while (spalte>=16) { spalte-=16; zeile++; }
	#endif
	#ifdef ONE_LINES
		zeile=0;
	#endif
	#ifdef TWO_LINES 
		#if !LCD_SER
			while (zeile>1) {lcd_scroll();zeile--;}
		#else
			if (zeile>1) {lcd_clear(); zeile=0;}
		#endif
	#endif
	#ifdef FOUR_LINES 
		#if !LCD_SER
			while (zeile>3) {lcd_scroll();zeile--;}
		#else
			if (zeile>3) {lcd_clear(); zeile=0;}
		#endif
	#endif
		
	lcd_cx = spalte;
	lcd_cy = zeile;
	
	//Berechnet Adresse fr die Zeile und schreibt sie ins DD-Ram
	zeile = LINE_ADDRESS[zeile];
	zeile += spalte;
	lcd_write(zeile,0);
}

#if !USE_LCD_STREAM
//------------------------------------------------------------------------------
void lcd_print_P (unsigned char zeile,unsigned char spalte,const char *Buffer,...)
{
	va_list ap;
	va_start (ap, Buffer);	
	
	int format_flag;
	char str_buffer[10];
	char str_null_buffer[10];
	char move = 0;
	char Base = 0;
	int tmp = 0;
	char by;
	char *ptr;
	
	lcd_pos(zeile,spalte);
	
	//Ausgabe der Zeichen
    for(;;)
	{
		by = pgm_read_byte(Buffer++);
		if(by==0) break; // end of format string
		
		if (by == '%')
		{
            by = pgm_read_byte(Buffer++);
			if (isdigit(by)>0)
				{
                                 
 				str_null_buffer[0] = by;
				str_null_buffer[1] = '\0';
				move = atoi(str_null_buffer);
                by = pgm_read_byte(Buffer++);
				}
				
			switch (by)
				{
				case 's':
                    ptr = va_arg(ap,char *);
                    while(*ptr) {lcd_write(*ptr++,1); }
                    break;
				case 'b':
					Base = 2;
					goto ConversionLoop;
				case 'c':
					//Int to char
					format_flag = va_arg(ap,int);
					lcd_write (format_flag++,1); 
					break;
				case 'i':
					Base = 10;
					goto ConversionLoop;
				case 'o':
					Base = 8;
					goto ConversionLoop;
				case 'x':
					Base = 16;
					//****************************
					ConversionLoop:
					//****************************
					itoa(va_arg(ap,int),str_buffer,Base);
					int b=0;
					while (str_buffer[b++] != 0){};
					b--;
					if (b<move)
						{
						move -=b;
						for (tmp = 0;tmp<move;tmp++)
							{
							str_null_buffer[tmp] = '0';
							}
						//tmp ++;
						str_null_buffer[tmp] = '\0';
						strcat(str_null_buffer,str_buffer);
						strcpy(str_buffer,str_null_buffer);
						}
					lcd_print_str (str_buffer);
					move =0;
					break;
				}
			
			}	
		else
			{
			lcd_print_char (by);
			} 
		} 
	va_end(ap);
}

//------------------------------------------------------------------------------
void lcd_print_str (char *Buffer)
{
	while (*Buffer != 0)
		{
		lcd_print_char (*Buffer++);
		};
};
#else
	//This display supports libc printf with scrolling, etc.
	int lcd_putchar(char c, FILE *stream) {
		if (c==0x0a) { lcd_cy++; c=0x0d; }
		if (c==0x0d) lcd_cx=0;
		else {
			lcd_pos(lcd_cy,lcd_cx);
			lcd_print_char(c);
			lcd_cx++;
		}
		return 0;
	}
	FILE lcd_stream = FDEV_SETUP_STREAM(lcd_putchar, NULL, _FDEV_SETUP_WRITE);
#endif

#endif //USE_LCD


